//UnixShell.c
//Written by Andy Schott, 6/2/99
//Developed using MS Visual C++ 5.0, Service Pack 1
//Implements a simple UnixShell that runs over Dos, translating
//some Unix commands to their Dos equivelants.

#include <ctype.h> //for toupper()
#include <direct.h> //for directory changing commands
#include <malloc.h> //to allocate memory for strings
#include <stdio.h> //basic input and output
#include <stdlib.h> //some other nifty commands
#include <string.h> //string handling

void addCwd(char *cmd,char *cwd);
int getDriveNum(char drive);
void headCmd(char *cmd);
int isCustomCmd(char *cmd);
void replace(char *str,char *add);
void setCwd(char *cmd,char *cwd);
int strStartsWith(char *cmd,char *str);
void tailCmd(char *cmd);
void unixFilter(char *cmd);

void main(int argc, char *argv[], char *envp[])
//argc - number of params passed to program
//argv - array of null-terminated params
//envp - array of null-terminated environment params
{
	char *cmd = malloc(96);
	char *cwd = malloc(96);
	char *temp,*lastSlash,*oldCwd;
	int error = -1;
	int curDrive = 3;
	int tempDrive = 0;

	printf("\nNow entering the UnixShell\n");

	//did use specify where to start?
	if(argc > 1)
	{
		error = chdir(argv[1]);
		if(error != 0)
		{
			printf("Bad directory name as parameter\n");
			chdir(getenv("winbootdir"));
		}
		_getdcwd(curDrive,cwd,96);
	}
	else
	{
		chdir(getenv("winbootdir"));
		_getdcwd(curDrive,cwd,96);
	}

	do
	{
		error = -1;
		printf("%s>",cwd);
		gets(cmd);
		if(strcmp(cmd,"exit") != 0)
		{
			//if it's a cd w/ no params, let it pass
			if(strcmp(cmd,"cd") == 0)
				system(cmd);
			//change the cwd
			else if(cmd[0] == 'c' && cmd[1] == 'd')
			{
				//was there a space after the cd?
				if(cmd[2] == ' ')
				{
					//change the directory, and where we think
					//the user currently is
					error = chdir(&cmd[3]);
					temp = strchr(&cmd[3],' ');
					if(error != 0 && temp == NULL)
						printf("Bad directory name\n");
					else
					{
						if(temp != NULL)
						{
							//change the directory correctly
							//if a space is in the name
							//some pretty wacky stuff
							setCwd(&cmd[3],cwd);
							//remember old cwd
							oldCwd = malloc(96);
							strcpy(oldCwd,cwd);
							lastSlash = strchr(cwd,'\\');
							while(lastSlash != NULL)
							{
								temp = lastSlash;
								temp = &temp[1];
								lastSlash = strchr(temp,'\\');
							}
							if(temp != NULL)
								lastSlash = temp;
							else
								lastSlash = cwd;
							lastSlash[6] = '\0';
							strcat(lastSlash,"~1");
							chdir(cwd);
							free(cwd);
							cwd = oldCwd;
						}
						else
							_getdcwd(curDrive,cwd,96);
					}
				}
				else
				{
					//change the directory, and where we think
					//the user currently is
					error = chdir(&cmd[2]);
					if(error != 0 && strchr(&cmd[3],' ') == NULL)
						printf("Bad directory name\n");
					else
					{
						if(strchr(&cmd[2],' ') != NULL)
							setCwd(&cmd[2],cwd);
						else
							_getdcwd(curDrive,cwd,96);
					}
				}
			}
			//change drives
			else if(cmd[1] == ':')
			{
				tempDrive = getDriveNum(cmd[0]);
				if(tempDrive == curDrive)
				{
					chdir(&cmd[2]);
					_getdcwd(curDrive,cwd,96);
				}
				else if(strlen(cmd) == 2)
				{
					curDrive = tempDrive;
					_chdrive(curDrive);
					_getdcwd(curDrive,cwd,96);
				}
				else
				{
					curDrive = tempDrive;
					_chdrive(curDrive);
					chdir(&cmd[2]);
					_getdcwd(curDrive,cwd,96);
				}
			}
			//make sure the user actually typed in a command
			else if(strcmp(cmd,"") != 0)
			{
				unixFilter(cmd); //get rid of unix commands
				//now, check if it is any of our commands
				if(!isCustomCmd(cmd))
				{
					//check if drive: is on end
					if(cmd[strlen(cmd)-1] == ':')
					{
						temp = malloc(96);
						strcpy(temp,cwd);
						_getdcwd(getDriveNum(cmd[strlen(cmd)-2]),cwd,96);
						cwd = &cwd[2];
						strcat(cmd,cwd);
						strcpy(cwd,temp);
						free(temp);
					}
#ifdef _DEBUG
					printf("%s\n",cmd);
#endif
					system(cmd);
					//add a \n at end of type,just in case
					if(strStartsWith(cmd,"type") || strStartsWith(cmd,"cat"))
						printf("\n");
				}
			}
		}
	}
	while(strcmp(cmd,"exit") != 0);

	//free up what we allocated, if needed
	if(cwd != NULL)
		free(cwd);
	if(cmd != NULL)
		free(cmd);

	printf("Now leaving the UnixShell\n");
}

int getDriveNum(char drive)
{
	return toupper(drive) - 64;
}

void setCwd(char *cmd,char *cwd)
//change where we think the user is
//only called if going to a directory with a space in the name
{
	//add a \ on end, if needed
	if(cwd[strlen(cwd)-1] != '\\')
		strcat(cwd,"\\");

	//get rid of leading and ending "
	if(cmd[0] == '\"')
	{
		cmd = &cmd[1];
		cmd[strlen(cmd)-1] = '\0';
	}
	//was drive letter specified first?
	if(cmd[1] == ':')
		strcpy(cwd,cmd);
	//absolute path?
	else if(cmd[0] == '\\')
		sprintf(cwd,"%c:%s",cwd[0],cmd);
	else
		strcat(cwd,cmd);
}

int isCustomCmd(char *cmd)
//is the command one we made?
{
	//the head command
	if(strStartsWith(cmd,"head"))
		headCmd(cmd);
	//the tail command
	else if(strStartsWith(cmd,"tail"))
		tailCmd(cmd);
	else
		return 0;

	return 1;
}

void headCmd(char *cmd)
//print out the first lines of a file
{
	int numLines,i;
	char *filename = &cmd[5];
	char *lines = malloc(4),*old;
	FILE *pFile;

	strcpy(lines,filename);
	old = strchr(filename,' ');
	//retrieve the command line params
	if(old != NULL)
	{
		old[0] = '\0';
		lines = strchr(lines,' ');
		lines = &lines[1];
		numLines = atoi(lines);
	}
	else
		numLines = 10;

	//now we just have to display the correct number of lines
	pFile = fopen(filename,"r");
	lines = malloc(25);
	old = lines;
	for(i = 0;i < numLines && old != NULL;i++)
	{
		do
		{
			old = fgets(lines,25,pFile);
			if(old == NULL)
				break;
			printf("%s",lines);
		}
		while(lines[strlen(lines)-1] != '\n');
	}
	free(lines);
	fclose(pFile);
}

void tailCmd(char *cmd)
//display the final lines of a file
{
	int numLines,i,fileSize;
	char *filename = &cmd[5];
	char *lines = malloc(4),*old;
	FILE *pFile;

	//determine the params
	strcpy(lines,filename);
	old = strchr(filename,' ');
	if(old != NULL)
	{
		old[0] = '\0';
		lines = strchr(lines,' ');
		lines = &lines[1];
		numLines = atoi(lines);
	}
	else
		numLines = 10;

	//get the number of lines in the file
	pFile = fopen(filename,"r");
	lines = malloc(25);
	fileSize = 0;
	do
	{
		do
		{
			old = fgets(lines,25,pFile);
		}
		while(lines[strlen(lines)-1] != '\n');
		if(old != NULL)
			fileSize++;
	}
	while(old != NULL);

	//we now have the number of lines in the file
	//we need to display the last numLines of them.
	fileSize -= numLines;
	if(fileSize <= 0) //can we just show it all?
	{
		sprintf(lines,"type %s",filename);
		system(lines);
		fclose(pFile);
		return;
	}

	fseek(pFile,0,SEEK_SET); //return to beginning
	for(i = 0;i < fileSize;i++)
	{
		do
		{
			fgets(lines,25,pFile);
		}
		while(lines[strlen(lines)-1] != '\n');
	}
	//now just display the rest of the file
	do
	{
		do
		{
			old = fgets(lines,25,pFile);
			if(old != NULL)
				printf(lines);
		}
		while(lines[strlen(lines)-1] != '\n');
	}
	while(old != NULL);

	free(lines);
	fclose(pFile);
}

void unixFilter(char *cmd)
/*Known Unix commands:
	ls
	rm
	mkdir
	rmdir
	cat
	mv
	cp
*/
{
	if(strStartsWith(cmd,"ls"))
		replace(cmd,"dir");
	else if(strStartsWith(cmd,"mkdir"))
		replace(cmd,"md");
	else if(strStartsWith(cmd,"rmdir"))
		replace(cmd,"rd");
	else if(strStartsWith(cmd,"rm"))
		replace(cmd,"del");
	else if(strStartsWith(cmd,"cat"))
		replace(cmd,"type");
	else if(strStartsWith(cmd,"mv"))
		replace(cmd,"move");
	else if(strStartsWith(cmd,"cp"))
		replace(cmd,"copy");
}

void replace(char *str,char *add)
//put in the correct Dos command for the Unix one
{
	char *temp = malloc(96),*param;
	param = strchr(str,' ');
	if(param != NULL)
	{
		sprintf(temp,"%s%s",add,param);
		strcpy(str,temp);
	}
	else
		strcpy(str,add);
}

int strStartsWith(char *cmd,char *str)
//does cmd start with str?
{
	unsigned int i;
	if(strlen(cmd) < strlen(str))
		return 0;
	for(i = 0;i < strlen(str);i++)
	{
		if(cmd[i] != str[i])
			return 0;
	}

	return 1;
}